@@ -36,7 +36,11 @@ logger = logging.getLogger(__name__)
 # endregion
 
 # region LocalImports
-from escriptorium_connector.utils import TimeoutWebsocket
+from escriptorium_connector.utils import (
+    TimeoutWebsocket,
+    get_all_forms,
+    get_form_details,
+)
 from escriptorium_connector.connector_errors import (
     EscriptoriumConnectorHttpError,
     EscriptoriumConnectorDtoSyntaxError,
@@ -763,6 +767,104 @@ class EscriptoriumConnector:
     def delete_line_type(self, line_type_pk: int) -> requests.Response:
         return self.__delete_url(f"{self.api_url}types/line/{line_type_pk}/")
 
+    # region document valid line types
+
+    # TODO: Warning, these functions are a hack to get past the lack of support
+    # in the API for creating document valid line types. They use a form POST
+    # to the eScriptorium website, since no API is available. They do NOT work
+    # if any annotation components, image annotations, or text annotations have
+    # already been created. The whole situation is quite poor right now and
+    # awaiting a fix.
+
+    def create_document_line_types_by_pk(self, doc_pk: int, line_type_pks: List[int]):
+        # Get the current ontology information
+        ontology_url = f"{self.base_url}document/{doc_pk}/ontology/"
+        forms = get_all_forms(
+            ontology_url,
+            self.http,
+        )
+        ontology_form = forms[0]
+        form_details = get_form_details(ontology_form)
+        data = {}
+        # Copy all existing data
+        for input_tag in form_details["inputs"]:
+            if input_tag["name"] not in data:
+                data[input_tag["name"]] = []
+            data[input_tag["name"]].append(input_tag["value"])
+
+        # Change only the valid region types
+        data["valid_line_types"] = data["valid_line_types"] + line_type_pks
+
+        res = self.http.post(
+            ontology_url,
+            data=data,
+            headers={**self.http.headers, "Referer": ontology_url},
+        )
+
+    def create_document_line_type_by_pk(self, doc_pk: int, line_type_pk: int):
+        self.create_document_line_types_by_pk(doc_pk, [line_type_pk])
+
+    def create_document_line_types(self, doc_pk: int, line_type_names: List[str]):
+        # Make sure all the desired line types exist and get their PKs
+        current_line_types = self.get_line_types().results
+        current_line_type_names = [x.name for x in current_line_types]
+        non_existant_line_types = [
+            x for x in line_type_names if x not in current_line_type_names
+        ]
+        line_type_pks = [x.pk for x in current_line_types if x.name in line_type_names]
+        for non_existant_line_type in non_existant_line_types:
+            new_line_type = self.create_line_type(
+                PostLineType(name=non_existant_line_type)
+            )
+            line_type_pks.append(new_line_type.pk)
+
+        self.create_document_line_types_by_pk(doc_pk, line_type_pks)
+
+    def create_document_line_type(self, doc_pk: int, line_type_name: str):
+        return self.create_document_line_types(doc_pk, [line_type_name])
+
+    def delete_document_line_types_by_pk(self, doc_pk: int, line_type_pks: List[int]):
+        # Get the current ontology information
+        ontology_url = f"{self.base_url}document/{doc_pk}/ontology/"
+        forms = get_all_forms(
+            ontology_url,
+            self.http,
+        )
+        ontology_form = forms[0]
+        form_details = get_form_details(ontology_form)
+        data = {}
+        # Copy all existing data
+        for input_tag in form_details["inputs"]:
+            if input_tag["name"] not in data:
+                data[input_tag["name"]] = []
+            data[input_tag["name"]].append(input_tag["value"])
+
+        # Change only the valid region types
+        data["valid_line_types"] = [
+            x for x in data["valid_line_types"] if x not in line_type_pks
+        ]
+
+        res = self.http.post(
+            ontology_url,
+            data=data,
+            headers={**self.http.headers, "Referer": ontology_url},
+        )
+
+    def delete_document_line_type_by_pk(self, doc_pk: int, line_type_pk: int):
+        self.delete_document_line_types_by_pk(doc_pk, [line_type_pk])
+
+    def delete_document_line_types(self, doc_pk: int, line_type_names: List[str]):
+        # Get the PKs of all line types to be deleted
+        current_line_types = self.get_line_types().results
+        delete_pks = [x.pk for x in current_line_types if x.name in line_type_names]
+
+        self.delete_document_line_types_by_pk(doc_pk, delete_pks)
+
+    def delete_document_line_type(self, doc_pk: int, line_type_name: str):
+        self.delete_document_line_types(doc_pk, [line_type_name])
+
+    # endregion
+
     # endregion
 
     # region Region API
@@ -797,8 +899,8 @@ class EscriptoriumConnector:
             f"{self.api_url}types/block/", GetRegionTypes
         )
 
-    def get_document_region_types(self, document_pk: int) -> List[GetRegionType]:
-        doc_data = self.get_document(document_pk)
+    def get_document_region_types(self, doc_pk: int) -> List[GetRegionType]:
+        doc_data = self.get_document(doc_pk)
         return [x for x in doc_data.valid_block_types]
 
     def create_region_type(self, region_type: PostRegionType) -> GetRegionType:
@@ -806,6 +908,110 @@ class EscriptoriumConnector:
             f"{self.api_url}types/block/", asdict(region_type), GetRegionType
         )
 
+    # region document valid region types
+
+    # TODO: Warning, these functions are a hack to get past the lack of support
+    # in the API for creating document valid region types. They use a form POST
+    # to the eScriptorium website, since no API is available. They do NOT work
+    # if any annotation components, image annotations, or text annotations have
+    # already been created. The whole situation is quite poor right now and
+    # awaiting a fix.
+
+    def create_document_region_types_by_pk(
+        self, doc_pk: int, region_type_pks: List[int]
+    ):
+        # Get the current ontology information
+        ontology_url = f"{self.base_url}document/{doc_pk}/ontology/"
+        forms = get_all_forms(
+            ontology_url,
+            self.http,
+        )
+        ontology_form = forms[0]
+        form_details = get_form_details(ontology_form)
+        data = {}
+        # Copy all existing data
+        for input_tag in form_details["inputs"]:
+            if input_tag["name"] not in data:
+                data[input_tag["name"]] = []
+            data[input_tag["name"]].append(input_tag["value"])
+
+        # Change only the valid region types
+        data["valid_block_types"] = data["valid_block_types"] + region_type_pks
+
+        res = self.http.post(
+            ontology_url,
+            data=data,
+            headers={**self.http.headers, "Referer": ontology_url},
+        )
+
+    def create_document_region_type_by_pk(self, doc_pk: int, region_type_pk: int):
+        self.create_document_region_types_by_pk(doc_pk, [region_type_pk])
+
+    def create_document_region_types(self, doc_pk: int, region_type_names: List[str]):
+        # Make sure all the desired region types exist and get their PKs
+        current_region_types = self.get_line_types().results
+        current_region_type_names = [x.name for x in current_region_types]
+        non_existant_region_types = [
+            x for x in region_type_names if x not in current_region_type_names
+        ]
+        region_type_pks = [
+            x.pk for x in current_region_types if x.name in region_type_names
+        ]
+        for non_existant_region_type in non_existant_region_types:
+            new_region_type = self.create_line_type(
+                PostLineType(name=non_existant_region_type)
+            )
+            region_type_pks.append(new_region_type.pk)
+
+        self.create_document_region_types_by_pk(doc_pk, region_type_pks)
+
+    def create_document_region_type(self, doc_pk: int, region_type_pk: str):
+        return self.create_document_region_types(doc_pk, [region_type_pk])
+
+    def delete_document_region_types_by_pk(
+        self, doc_pk: int, region_type_pks: List[int]
+    ):
+        # Get the current ontology information
+        ontology_url = f"{self.base_url}document/{doc_pk}/ontology/"
+        forms = get_all_forms(
+            ontology_url,
+            self.http,
+        )
+        ontology_form = forms[0]
+        form_details = get_form_details(ontology_form)
+        data = {}
+        # Copy all existing data
+        for input_tag in form_details["inputs"]:
+            if input_tag["name"] not in data:
+                data[input_tag["name"]] = []
+            data[input_tag["name"]].append(input_tag["value"])
+
+        # Change only the valid region types
+        data["valid_block_types"] = [
+            x for x in data["valid_block_types"] if x not in region_type_pks
+        ]
+
+        res = self.http.post(
+            ontology_url,
+            data=data,
+            headers={**self.http.headers, "Referer": ontology_url},
+        )
+
+    def delete_document_region_type_by_pk(self, doc_pk: int, region_type_pk: int):
+        self.delete_document_region_types_by_pk(doc_pk, [region_type_pk])
+
+    def delete_document_region_types(self, doc_pk: int, region_type_names: List[str]):
+        # Get the PKs of all region types to be deleted
+        current_region_types = self.get_region_types().results
+        delete_pks = [x.pk for x in current_region_types if x.name in region_type_names]
+
+        self.delete_document_region_types_by_pk(doc_pk, delete_pks)
+
+    def delete_document_region_type(self, doc_pk: int, region_type_name: str):
+        self.delete_document_region_types(doc_pk, [region_type_name])
+
+    # endregion
+
     # endregion
 
     # region Transcription API
